#include "TrackingAction.hh"
#include "DetectorConstruction.hh"
#include "G4RunManager.hh"
#include "G4ParticleDefinition.hh"
#include "G4DynamicParticle.hh"
#include "G4Track.hh"
#include "G4ThreeVector.hh"
#include "G4SystemOfUnits.hh"
#include "Randomize.hh"


TrackingAction::TrackingAction() : G4UserTrackingAction()
{;}


TrackingAction::~TrackingAction()
{;}


void TrackingAction::PreUserTrackingAction(const G4Track* aTrack)
{
  // Decision is taken to track a primary only if it will hit the detector, allowing for multiple scattering

  // Select only for primary particles
  if (aTrack->GetParentID() != 0) {
    return;
  }

  // Check if primary track will hit the detector

  G4PhysicalVolumeStore * store = G4PhysicalVolumeStore::GetInstance();
  G4VPhysicalVolume * physDet = store->GetVolume("Detector");
  G4VPhysicalVolume * physEarth = store->GetVolume("Earth");

  G4Box * box  = dynamic_cast<G4Box*>(physDet->GetLogicalVolume()->GetSolid());
  G4double x = box->GetXHalfLength();
  G4double y = box->GetYHalfLength();
  G4double detRadius = sqrt(x*x + y*y);

  G4ThreeVector tmp = physDet->GetObjectTranslation() /*+ physEarth->GetObjectTranslation()*/ ;
//  HARDCODED detector z. To be CHECKED !!!
  G4ThreeVector detPosition = G4ThreeVector(tmp.x(),tmp.y(),-19.0*m); 

  G4double kineticE = aTrack->GetKineticEnergy();

  const G4ParticleDefinition * partDef = aTrack->GetParticleDefinition();

  // Compute the multiplescatttering rms (from PDG)
  G4Material * earth = physEarth->GetLogicalVolume()->GetMaterial();
  const G4double X0 = earth->GetRadlen();
  const G4double mass = partDef->GetPDGMass();
  const G4double charge = partDef->GetPDGCharge();
  const G4double charge2 = charge*charge;

  // Kinematical variables
  const G4double energy = kineticE + mass;
  const G4double energy2 = energy*energy;
  const G4double momentum2 = energy2 - mass*mass;
  const G4double betapc = momentum2/energy;
  const G4double beta2 = momentum2/energy2;

  const G4ThreeVector pos = aTrack->GetPosition();
  const G4ThreeVector dir = aTrack->GetMomentumDirection();

// Compute track position at detector

  // lambda is the distance along the track direction from its original height to the detector's height
  const G4double lambda = (detPosition.z()-pos.z())*dir.mag()/dir.z();

//  G4cout << "Pos z/Det z/lambda (m) : " << pos.z()/m << "    " << detPosition.z()/m << "   " << lambda/m << G4endl;

  const G4ThreeVector newpos = pos + lambda * dir;

// Compute multiple scattering 
  const G4double depth = abs(detPosition.z());
  const G4double slantDepth = abs(depth/cos(dir.theta()));

  G4double theta0; 
  if (charge  == 0.0) {
    theta0 = 0.0;
  }
  else {
    theta0 = 13.6*MeV/betapc * abs(charge) * sqrt(slantDepth/X0) * (1.0 + 0.038*log(slantDepth/X0*charge2/beta2)); 
  }

  // Deviation due to multiple scattering; TO BE CHECKED !!! In any case for now R0 yileds typically < 1 m, so no big impact considering  
  // the safety factor of 5 x detRadius
  const G4double R0 = slantDepth * tan(theta0);

  if ( (newpos-detPosition).perp() > 3.0*(detRadius+R0) ) {      //factor 3 (but above is safety factor 5(?))
    fpTrackingManager->GetTrack()->SetTrackStatus(fStopAndKill);
  } 
  else {
    // if ( 180.*deg - dir.theta() > 45.0*deg)
    // G4cout << (180.0*deg-dir.theta())/deg << "   " << pos.perp()/m << G4endl;
  }

  // G4cout << X0/cm << "   " << slantDepth/m << "  " << mass/MeV << "   " << beta2 << "   " << theta0*180.0/CLHEP::pi << "   " << R0/m << G4endl;    
  // G4cout << "Radius at detector before/after MS :  " << (newpos-detPosition).perp()/m  << "   " <<  R0/m << G4endl;

  return;
}


void TrackingAction::PostUserTrackingAction(const G4Track*)
{;}
